/*
 * This file is part of Monitor VPN.
 *
 * Copyright © 2022 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

namespace MonitorVPN {

    /**
     * Class to monitor the VPN connection.
     */
    public class Monitor : Object {

        /* One million. */
        private const double ONE_MILLION = 1000000;
        /* Config file. */
        private const string UUID_PATH = "/etc/monitor-vpn/monitor-vpn.conf";

        /* Keep going flag. */
        private bool keep_going;
        /* The connection UUID. */
        private string uuid;
        /* The main loop. */
        private GLib.MainLoop loop;
        /* The last monitoring datetime. */
        private GLib.DateTime last;

        /* Parses the UUID from the configuration file. */
        private  string get_uuid() throws GLib.Error {
            var file = GLib.File.new_for_path(UUID_PATH);
            if (!file.query_exists()) {
                stderr.printf("File '%s' doesn't exist.\n", file.get_path());
                throw new GLib.Error(Quark.from_string("mvpn"),
                                     ErrorCode.IRRECOVERABLE,
                                     "Missing configuration file");
            }

            var dis = new DataInputStream(file.read());
            string line;
            while ((line = dis.read_line(null)) != null) {
                line = line.strip();
                if (!line.has_prefix("VPN_UUID="))
                    continue;
                var t = line.split("=");
                return t[1].replace("\"", "");
            }
            throw new GLib.Error(Quark.from_string("mvpn"),
                                 ErrorCode.IRRECOVERABLE,
                                 "Invalid configuration file");
        }

        /**
         * Sets the initial state of the monitor.
         * @throws GLib.Error if the UUID cannot be obtained.
         */
        public Monitor() throws GLib.Error {
            loop = new GLib.MainLoop();
            uuid = get_uuid();
        }

        /* The monitoring idle source function. */
        private bool do_monitoring() {
            if (!keep_going) {
                GLib.Idle.add(() => {
                        loop.quit();
                        return GLib.Source.REMOVE;
                    });
                return GLib.Source.REMOVE;
            }
            GLib.Thread.usleep((ulong)(ONE_MILLION / 5));
            var now = new GLib.DateTime.now_utc();
            if (now.difference(last) >= ONE_MILLION)
                return GLib.Source.CONTINUE;
            try {
                var nm = new NetworkManager();
                if (nm.check_connectivity() != ConnectivityState.FULL)
                    return GLib.Source.CONTINUE;
                try {
                    var avpn = nm.get_connection_by_uuid(uuid);
                    var s = avpn.get_state();
                    if (s == ActiveConnectionState.ACTIVATED ||
                        s == ActiveConnectionState.ACTIVATING)
                        return GLib.Source.CONTINUE;
                } catch (NMError nme) {
                    return GLib.Source.CONTINUE;
                }
                var settings = new Settings();
                var conn = settings.get_connection_by_uuid(uuid);
                var pc = nm.get_primary_connection();
                stderr.printf("Activating %s...\n", conn);
                nm.activate_connection(conn, pc);
                stderr.printf("Connection %s activated.\n", conn);
            } catch (GLib.Error e) {
                stderr.printf("%d: %s\n", e.code, e.message);
                keep_going = false;
                loop.quit();
                return GLib.Source.REMOVE;
            }
            last = now;
            return GLib.Source.CONTINUE;
        }

        /* Ends the monitoring. */
        private bool handle_sigterm() {
            keep_going = false;
            stderr.printf("Terminating monitoring...\n");
            return GLib.Source.REMOVE;
        }

        /**
         * Starts the monitoring.
         */
        public void start() {
            keep_going = true;
            last = new GLib.DateTime.now_utc();
            GLib.Unix.signal_add(GLib.ProcessSignal.TERM, handle_sigterm);
            GLib.Idle.add(do_monitoring);
            stderr.printf("Monitoring starting...\n");
            loop.run();
            stderr.printf("Monitoring terminated.\n");
        }
    }
}
