/*
 * This file is part of Monitor VPN.
 *
 * Copyright © 2022 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

namespace MonitorVPN {

    /**
     * Error codes.
     */
    public enum ErrorCode {
        /**
         * Irrecoverable error.
         */
        IRRECOVERABLE = 123;
    }

    /**
     * Error domain for NetworkManager
     */
    public errordomain NMError {
        /**
         * No such UUID.
         */
        NO_SUCH_UUID;
    }

    /**
     * Connectivity states.
     */
    public enum ConnectivityState {
        /**
         * Network connectivity is unknown. This means the connectivity checks
         * are disabled (e.g. on server installations) or has not run yet. The
         * graphical shell should assume the Internet connection might be
         * available and not present a captive portal window.
         */
        UNKNOWN = 0,

        /**
         * The host is not connected to any network. There's no active
         * connection that contains a default route to the internet and thus it
         * makes no sense to even attempt a connectivity check. The graphical
         * shell should use this state to indicate the network connection is
         * unavailable.
         */
        NONE = 1,

        /**
         * The Internet connection is hijacked by a captive portal gateway. The
         * graphical shell may open a sandboxed web browser window (because the
         * captive portals typically attempt a man-in-the-middle attacks against
         * the https connections) for the purpose of authenticating to a gateway
         * and retrigger the connectivity check with CheckConnectivity() when
         * the browser window is dismissed.
         */
        PORTAL = 2,

        /**
         * The host is connected to a network, does not appear to be able to
         * reach the full Internet, but a captive portal has not been detected.
         */
        LIMITED = 3,

        /**
         * The host is connected to a network, and appears to be able to reach
         * the full Internet.
         */
        FULL = 4;

        /**
         * Returns the connectivity state from a uint.
         * @param uuid the uuid.
         * @return the connectivity state from a uint.
         */
        public static ConnectivityState from_uint(uint uuid) {
            if (uuid >= UNKNOWN && uuid <= FULL)
                return (ConnectivityState)uuid;
            return UNKNOWN;
        }
    }

    /**
     * ActiveConnectionState values indicate the state of a connection to a
     * specific network while it is starting, connected, or disconnecting from
     * that network.
     */
    public enum ActiveConnectionState {

        /**
         * The state of the connection is unknown.
         */
        UNKNOWN = 0,

        /**
         * A network connection is being prepared.
         */
        ACTIVATING = 1,

        /**
         * There is a connection to the network.
         */
        ACTIVATED = 2,

        /**
         * The network connection is being torn down and cleaned up.
         */
        DEACTIVATING = 3,

        /**
         * The network connection is disconnected and will be removed.
         */
        DEACTIVATED = 4;

        /**
         * Returns the action connectiion state from a uint.
         * @param uuid the uuid.
         * @return the connectivity state from a uint.
         */
        public static ActiveConnectionState from_uint(uint uuid) {
            if (uuid >= UNKNOWN && uuid <= DEACTIVATED)
                return (ActiveConnectionState)uuid;
            return UNKNOWN;
        }
    }

    /**
     * The Properties proxy.
     */
    [DBus (name = "org.freedesktop.DBus.Properties")]
    public interface PropertiesProxy : GLib.Object {

        /**
         * Get the value for the given property.
         * @param if_name The interface this property is associated with.
         * @param prop_name The name of the property.
         * @throws GLib.Error if something goes wrong.
         */
        public abstract GLib.Variant get(string if_name, string prop_name)
            throws GLib.Error;
    }

    /**
     * The NetworkManager proxy.
     */
    [DBus (name = "org.freedesktop.NetworkManager")]
    public interface NetworkManagerProxy : GLib.Object, PropertiesProxy {

        /**
         * Re-check the network connectivity state.
         * @return the current connectivity state.
         * @throws GLib.Error if something goes wrong.
         */
        public abstract uint check_connectivity() throws GLib.Error;

        /**
         * Activate a connection using the supplied device.
         * @param connection The connection to activate. If "/" is given, a
         *                   valid device path must be given, and NetworkManager
         *                   picks the best connection to activate for the given
         *                   device. VPN connections must always pass a valid
         *                   connection path.
         * @param device The object path of device to be activated for physical
         *               connections. This parameter is ignored for VPN
         *               connections, because the specific_object (if provided)
         *               specifies the device to use.
         * @param specific_object The path of a connection-type-specific object
         *                        this activation should use. This parameter is
         *                        currently ignored for wired and mobile
         *                        broadband connections, and the value of "/"
         *                        should be used (ie, no specific object). For
         *                        Wi-Fi connections, pass the object path of a
         *                        specific AP from the card's scan list, or "/"
         *                        to pick an AP automatically. For VPN
         *                        connections, pass the object path of an
         *                        ActiveConnection object that should serve as
         *                        the "base" connection (to which the VPN
         *                        connections lifetime will be tied), or pass
         *                        "/" and NM will automatically use the current
         *                        default device.
         * @throws GLib.Error if something goes wrong.
         */
        public abstract GLib.ObjectPath activate_connection(
            GLib.ObjectPath connection,
            GLib.ObjectPath device,
            GLib.ObjectPath specific_object) throws GLib.Error;
    }

    /**
     * Settings proxy.
     */
    [DBus (name = "org.freedesktop.NetworkManager.Settings")]
    public interface SettingsProxy : GLib.Object, PropertiesProxy {

        /**
         * List the saved network connections known to NetworkManager.
         * @return the list of connections.
         * @throws GLib.Error if something goes wrong.
         */
        public abstract GLib.ObjectPath[] list_connections() throws GLib.Error;

        /**
         * Retrieve the object path of a connection, given that connection's UUID.
         * @param uuid The UUID to find the connection object path for.
         * @return the connection's object path.
         * @throws GLib.Error if something goes wrong.
         */
        public abstract GLib.ObjectPath get_connection_by_uuid(string uuid)
            throws GLib.Error;
    }

    /**
     * Connection proxy.
     */
    [DBus (name = "org.freedesktop.NetworkManager.Settings.Connection")]
    public interface ConnectionProxy : GLib.Object, PropertiesProxy {

        /**
         * Get the settings maps describing this network configuration. This
         * will never include any secrets required for connection to the
         * network, as those are often protected. Secrets must be requested
         * separately using the get_secrets() method.
         * @return the nested settings maps describing this object.
         * @throws GLib.Error if something goes wrong.
         */
        public abstract
            GLib.HashTable<string, GLib.HashTable<string, GLib.Variant>>
            get_settings() throws GLib.Error;
    }

    /**
     * Active proxy.
     */
    [DBus (name = "org.freedesktop.NetworkManager.Connection.Active")]
    public interface ActiveProxy : GLib.Object, PropertiesProxy {}

    /**
     * Represents an attempt to connect to a network using the details provided
     * by a Connection object. The Active object tracks the life-cycle of the
     * connection attempt and if successful indicates whether the connected
     * network is the "default" or preferred network for access. NetworkManager
     * has the concept of connections, which can be thought of as settings, a
     * profile or a configuration that can be applied on a networking
     * device. Such settings-connections are exposed as objects and the
     * active-connection expresses this relationship between device and
     * settings-connection. At any time a settings-connection can only be
     * activated on one device and vice versa. However, during activation and
     * deactivation multiple active-connections can reference the same device or
     * settings-connection as they are waiting to be activated or to be
     * deactivated.
     */
    public class Active : GLib.Object {

        /* The proxy object. */
        public ActiveProxy active;

        /**
         * Sets the initial state of the Active object.
         * @param path the object path for the active connection.
         * @throws GLib.Error if something goes wrong.
         */
        public Active(string path) throws GLib.Error {
            active = Bus.get_proxy_sync(GLib.BusType.SYSTEM,
                                        "org.freedesktop.NetworkManager",
                                        path);
        }

        /**
         * Returns the UUID of the Active object.
         * @return the UUID of the Active object.
         * @throws GLib.Error if something goes wrong.
         */
        public string get_uuid() throws GLib.Error {
            var gv =  active.get(
                "org.freedesktop.NetworkManager.Connection.Active", "Uuid");
            return gv.get_string();
        }

        /**
         * Returns the state of the Active object.
         * @return the state of the Active object.
         * @throws GLib.Error if something goes wrong.
         */
        public ActiveConnectionState get_state() throws GLib.Error {
            var gv = active.get(
                "org.freedesktop.NetworkManager.Connection.Active", "State");
            return gv.get_uint32();
        }

        /**
         * Returns the connection path of the Active object.
         * @return the connection path of the Active object.
         * @throws GLib.Error if something goes wrong.
         */
        public string get_connection_path() throws GLib.Error {
            var gv = active.get(
                "org.freedesktop.NetworkManager.Connection.Active",
                "Connection");
            return gv.get_string();
        }

        /**
         * Returns the specific object path of the Active object.
         * @return the specific object path of the Active object.
         * @throws GLib.Error if something goes wrong.
         */
        public string get_specific_object_path() throws GLib.Error {
            var gv = active.get(
                "org.freedesktop.NetworkManager.Connection.Active",
                "SpecificObject");
            return gv.get_string();
        }
    }

    /**
     * Connection Manager.
     */
    public class NetworkManager : GLib.Object {

        /* The proxy. */
        private NetworkManagerProxy nm;

        /**
         * Sets the initial state of the NetworkManager object.
         * @throws GLib.Error if something goes wrong.
         */
        public NetworkManager() throws GLib.Error {
            nm = Bus.get_proxy_sync(GLib.BusType.SYSTEM,
                                    "org.freedesktop.NetworkManager",
                                    "/org/freedesktop/NetworkManager");
        }

        /**
         * Returns the connectivity state of the NetworkManager object.
         * @return the connectivity state of the NetworkManager object.
         * @throws GLib.Error if something goes wrong.
         */
        public ConnectivityState check_connectivity() throws GLib.Error {
            return nm.check_connectivity();
        }

        /**
         * Returns the active connections state of the NetworkManager object.
         * @return the active connections state of the NetworkManager object.
         * @throws GLib.Error if something goes wrong.
         */
        public Active[] active_connections() throws GLib.Error {
            var acs = nm.get("org.freedesktop.NetworkManager",
                             "ActiveConnections");
            var objv = acs.get_objv();
            Active[] active_connections = new Active[objv.length];
            int i = 0;
            foreach (var path in objv) {
                active_connections[i++] = new Active(path);
            }
            return active_connections;
        }

        /**
         * Returns an active connection by UUID.
         * @param uuid the UUID.
         * @return an active connection by UUID.
         * @throws NMError.NO_SUCH_UUID if there is no active connection with
         *         such an UUID.
         * @throws GLib.Error if something else goes wrong.
         */
        public Active get_connection_by_uuid(string uuid)
            throws NMError, GLib.Error {
            var acs = active_connections();
            for (int i = 0; i < acs.length; i++) {
                var ac = acs[i];
                var _uuid = ac.get_uuid();
                if (uuid == _uuid)
                    return ac;
            }
            throw new NMError.NO_SUCH_UUID("No such UUID");
        }

        /**
         * Returns the primary connection of the NetworkManager object.
         * @return the primary connection of the NetworkManager object.
         * @throws GLib.Error if something goes wrong.
         */
        public string get_primary_connection() throws GLib.Error {
            var gv = nm.get("org.freedesktop.NetworkManager",
                            "PrimaryConnection");
            return gv.get_string();
        }

        /**
         * Activates a connection.
         * @param connection the path of the connection to activate.
         * @param specific_object the specific object path.
         * @throws GLib.Error if something goes wrong.
         */
        public string activate_connection(string connection,
                                          string specific_object)
            throws GLib.Error {
            return nm.activate_connection(new GLib.ObjectPath(connection),
                                          new GLib.ObjectPath("/"),
                                          new GLib.ObjectPath(specific_object));
        }
    }

    /**
     * Settings allows clients to view and administrate the connections stored
     * and used by NetworkManager.
     */
    public class Settings : GLib.Object {

        /* The proxy. */
        private SettingsProxy settings;

        /**
         * Sets the initial state of the Settings object.
         * @throws GLib.Error if something goes wrong.
         */
        public Settings() throws GLib.Error {
            settings =
                Bus.get_proxy_sync(GLib.BusType.SYSTEM,
                                   "org.freedesktop.NetworkManager",
                                   "/org/freedesktop/NetworkManager/Settings");
        }

        /**
         * Retrieve the object path of a connection, given that connection's
         * UUID.
         * @param uuid the connection's UUDI.
         * @return the object path of a connection.
         */
        public string get_connection_by_uuid(string uuid) throws GLib.Error {
            return settings.get_connection_by_uuid(uuid);
        }
    }
}
